// ========== begin copyright notice ==============
// This file is part of 
// ---------------
// xaifBooster
// ---------------
// Distributed under the BSD license as follows:
// Copyright (c) 2005, The University of Chicago
// All rights reserved.
//
// Redistribution and use in source and binary forms, 
// with or without modification, are permitted provided that the following conditions are met:
//
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright notice, 
//      this list of conditions and the following disclaimer in the documentation 
//      and/or other materials provided with the distribution.
//    - Neither the name of The University of Chicago nor the names of its contributors 
//      may be used to endorse or promote products derived from this software without 
//      specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
// SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// General Information:
// xaifBooster is intended for the transformation of 
// numerical programs represented as xml files according 
// to the XAIF schema. It is part of the OpenAD framework. 
// The main application is automatic 
// differentiation, i.e. the generation of code for 
// the computation of derivatives. 
// The following people are the principal authors of the 
// current version: 
// 	Uwe Naumann
//	Jean Utke
// Additional contributors are: 
//	Andrew Lyons
//	Peter Fine
//
// For more details about xaifBooster and its use in OpenAD please visit:
//   http://www.mcs.anl.gov/openad
//
// This work is partially supported by:
// 	NSF-ITR grant OCE-0205590
// ========== end copyright notice ==============
#include "xaifBooster/utils/inc/PrintManager.hpp"

#include "xaifBooster/system/inc/BasicBlock.hpp"
#include "xaifBooster/system/inc/Assignment.hpp"
#include "xaifBooster/system/inc/Scope.hpp"
#include "xaifBooster/system/inc/SubroutineCall.hpp"
#include "xaifBooster/system/inc/VariableSymbolReference.hpp"
#include "xaifBooster/system/inc/ArrayAccess.hpp"
#include "xaifBooster/system/inc/Argument.hpp"
#include "xaifBooster/system/inc/ConceptuallyStaticInstances.hpp"
#include "xaifBooster/system/inc/CallGraph.hpp"

#include "xaifBooster/algorithms/DerivativePropagator/inc/DerivativePropagatorSaxpy.hpp"

#include "xaifBooster/algorithms/InlinableXMLRepresentation/inc/InlinableSubroutineCall.hpp"

#include "xaifBooster/algorithms/TypeChange/inc/BasicBlockAlgParameter.hpp"

#include "xaifBooster/algorithms/AdjointUtils/inc/BasicBlockPrintVersion.hpp"

#include "xaifBooster/algorithms/BasicBlockPreaccumulationTapeAdjoint/inc/BasicBlockAlg.hpp"
#include "xaifBooster/algorithms/BasicBlockPreaccumulationTapeAdjoint/inc/BasicBlockElementAlg.hpp"

using namespace xaifBooster;

namespace xaifBoosterBasicBlockPreaccumulationTapeAdjoint { 

  BasicBlockAlg::BasicBlockAlg(BasicBlock& theContaining) : 
    xaifBooster::BasicBlockAlgBase(theContaining),
    xaifBoosterBasicBlockPreaccumulation::BasicBlockAlg(theContaining) { 
  }

  BasicBlockAlg::~BasicBlockAlg() { 
    for (PlainBasicBlock::BasicBlockElementList::const_iterator li=myBasicBlockElementListAnonymousReversal.begin();
         li!=myBasicBlockElementListAnonymousReversal.end();
         li++) { 
      if (*li)
	delete *li;
    } 
    for (PlainBasicBlock::BasicBlockElementList::const_iterator li=myBasicBlockElementListAnonymousReversal.begin();
         li!=myBasicBlockElementListExplicitReversal.end();
         li++) { 
      if (*li)
	delete *li;
    } 
  } 

  void
  BasicBlockAlg::printXMLHierarchy(std::ostream& os) const { 
    PrintManager& pm=PrintManager::getInstance();
    os << pm.indent() 
       << "<"
       << PlainBasicBlock::ourXAIFName.c_str()
       << " " 
       << PlainBasicBlock::our_myId_XAIFName.c_str()
       << "=\"" 
       << getContaining().getId().c_str()
       << "\" " 
       << ObjectWithAnnotation::our_myAnnotation_XAIFName.c_str() 
       << "=\""
       << getContaining().getAnnotation().c_str()
       << "\" " 
       << PlainBasicBlock::our_myScopeId_XAIFName.c_str() 
       << "=\""
       << getContaining().getScope().getId().c_str()
       << "\">" 
       << std::endl;
    const PlainBasicBlock::BasicBlockElementList& aBasicBlockElementList(getBasicBlockElementList(xaifBoosterAdjointUtils::BasicBlockPrintVersion::get()));
    for (PlainBasicBlock::BasicBlockElementList::const_iterator li=aBasicBlockElementList.begin();
	 li!=aBasicBlockElementList.end();
	 li++)
      (*(li))->printXMLHierarchy(os);
    os << pm.indent()
       << "</"
       << PlainBasicBlock::ourXAIFName.c_str()
       << ">"
       << std::endl;
    pm.releaseInstance();
  } // end of BasicBlockAlg::printXMLHierarchy
  
  std::string BasicBlockAlg::debug () const { 
    std::ostringstream out;
    out << "xaifBoosterBasicBlockPreaccumulationTapeAdjoint::BasicBlockAlg[" << this
	<< "]" << std::ends;  
    return out.str();
  } // end of BasicBlockAlg::debug

  void BasicBlockAlg::traverseToChildren(const GenericAction::GenericAction_E anAction_c) { 
  } 

  xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& 
  BasicBlockAlg::addInlinableSubroutineCall(const std::string& aSubroutineName,
					    const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall* aNewCall_p(new xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall(aSubroutineName));
    getBasicBlockElementList(aReversalType).push_back(aNewCall_p);
    return *aNewCall_p;									     
  } 
  
  SubroutineCall& 
  BasicBlockAlg::addSubroutineCall(const Symbol& aSubroutineNameSymbol,
				   const Scope& aSubroutineNameScope,
				   ActiveUseType::ActiveUseType_E anActiveUse,
				   const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    SubroutineCall* aNewCall_p(new SubroutineCall(aSubroutineNameSymbol,
						  aSubroutineNameScope,
						  anActiveUse,
						  true));
    aNewCall_p->setId("reverse_call");
    getBasicBlockElementList(aReversalType).push_back(aNewCall_p);
    return *aNewCall_p;									     
  } 

  const Assignment& 
  BasicBlockAlg::addConstantAssignment(const BaseConstant& theConstant,
				       const Symbol& aTemporarySymbol,
				       const Scope& theScopeOfTheTemporarySymbol,
				       const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    Assignment* theNewAssignment_p(new Assignment(false));
    theNewAssignment_p->setId("tape_adjoint_constant_assignment");
    getBasicBlockElementList(aReversalType).push_back(theNewAssignment_p);
    Constant* theConstantRHS_p(new Constant(theConstant.getType(),false));
    theConstantRHS_p->setFromString(theConstant.toString());
    theConstantRHS_p->setId(theNewAssignment_p->getRHS().getNextVertexId());
    theNewAssignment_p->getRHS().supplyAndAddVertexInstance(*theConstantRHS_p);
    // create a new symbol and add a new VariableSymbolReference in the Variable
    VariableSymbolReference* theNewVariableSymbolReference_p=
      new VariableSymbolReference(getContaining().getScope().
				  getSymbolTable().
				  addUniqueAuxSymbol(SymbolKind::VARIABLE,
						     SymbolType::INTEGER_STYPE,
						     SymbolShape::SCALAR,
						     false),
				  getContaining().getScope());
    theNewVariableSymbolReference_p->setId("1");
    theNewVariableSymbolReference_p->setAnnotation("xaifBoosterBasicBlockPreaccumulationTapeAdjoint::BasicBlockAlg::addConstantAssignment");
    // pass it on to the LHS and relinquish ownership
    theNewAssignment_p->getLHS().supplyAndAddVertexInstance(*theNewVariableSymbolReference_p);
    theNewAssignment_p->getLHS().getAliasMapKey().setTemporary();
    theNewAssignment_p->getLHS().getDuUdMapKey().setTemporary();
    return *theNewAssignment_p;									     
  } 

  void BasicBlockAlg::reinterpretArrayAccess(const Variable& theOriginalVariable,
					     Variable& theNewVariable,
					     const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    theOriginalVariable.copyMyselfInto(theNewVariable);
    if (!theOriginalVariable.hasArrayAccess() || aReversalType==ForLoopReversalType::EXPLICIT)
      return;
    // otherwise we will replace index expressions: 
    ArrayAccess::IndexTripletListType& theNewIndexTripletList(theNewVariable.getArrayAccess().getIndexTripletList());
    const ArrayAccess::IndexTripletListType& theOriginalIndexTripletList(theOriginalVariable.getArrayAccess().getIndexTripletList());
    ArrayAccess::IndexTripletListType::reverse_iterator aNewIndexTripletListTypeI=theNewIndexTripletList.rbegin() ;
    for (ArrayAccess::IndexTripletListType::const_reverse_iterator anOriginalIndexTripletListTypeCI=theOriginalIndexTripletList.rbegin();
	 anOriginalIndexTripletListTypeCI!=theOriginalIndexTripletList.rend();
	 ++anOriginalIndexTripletListTypeCI,
	   ++aNewIndexTripletListTypeI) { 
      for (IndexTriplet::IndexPairList::const_iterator anIndexPairListCI=(*anOriginalIndexTripletListTypeCI)->getIndexPairList().begin();
	   anIndexPairListCI!=(*anOriginalIndexTripletListTypeCI)->getIndexPairList().end();
	   ++anIndexPairListCI) { 
	// now we have two cases, essentially the expression is a single vertex with a constant 
	// (this discounts constant expressions, this is a todo which might be dealt with later or 
	// it may be completelt superceded by a TBR analysis)
	const Expression& theIndexExpression(*((*anIndexPairListCI).second));
	if (theIndexExpression.numVertices()==1
	    && 
	    !(*(theIndexExpression.vertices().first)).isArgument()) { 
	  // do nothing
	  continue;
	}
	// it is a variable or expression whose value we pushed and now want to pop
	// and replace. First clear the old index expression that we copied into aNewIndexTripletListTypeI
	(*aNewIndexTripletListTypeI)->getExpression((*anIndexPairListCI).first).clear();
	// pop the value
	xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& thePopCall(addInlinableSubroutineCall("pop_i",
													      aReversalType));
	thePopCall.setId("inline_pop_i");
	Variable& theInlineVariable(thePopCall.addConcreteArgument(1).getArgument().getVariable());
	// give it a name etc.
	// create a new symbol and add a new VariableSymbolReference in the Variable
	VariableSymbolReference* theNewVariableSymbolReference_p=
	  new VariableSymbolReference(getContaining().getScope().
				      getSymbolTable().
				      addUniqueAuxSymbol(SymbolKind::VARIABLE,
							 SymbolType::INTEGER_STYPE,
							 SymbolShape::SCALAR,
							 false),
				      getContaining().getScope());
	theNewVariableSymbolReference_p->setId("1");
	theNewVariableSymbolReference_p->setAnnotation("xaifBoosterBasicBlockPreaccumulationTapeAdjoint::BasicBlockAlg::reinterpretArrayAccess");
	// pass it on to the variable and relinquish ownership
	theInlineVariable.supplyAndAddVertexInstance(*theNewVariableSymbolReference_p);
	theInlineVariable.getAliasMapKey().setTemporary();
	theInlineVariable.getDuUdMapKey().setTemporary();
	// create a copy of the variable in the indexExpression: 
	Argument& theNewArgument(*new Argument);
	// relinquish ownership to the index expression: 
	(*aNewIndexTripletListTypeI)->getExpression((*anIndexPairListCI).first).supplyAndAddVertexInstance(theNewArgument);
	theNewArgument.setId(1);
	theInlineVariable.copyMyselfInto(theNewArgument.getVariable());
      } // end for loop through the pairlist
    } // end for i
  } 

  void BasicBlockAlg::algorithm_action_5() { // adjoin the DerivativePropagators
    if (getContaining().getBasicBlockElementList().empty())
      return;
    xaifBoosterTypeChange::BasicBlockAlgParameter::instance().set(*this); // in BasicBlockAlg::algorithm_action_5()
    // mesh the BasicBlockElements with the Sequences
    PlainBasicBlock::BasicBlockElementList::const_reverse_iterator aBasicBlockElementListRI=getContaining().getBasicBlockElementList().rbegin();
    SequencePList::const_reverse_iterator aSequencePListRI = myUniqueSequencePList.rbegin();
    SequencePList::const_reverse_iterator aSequencePListRend = myUniqueSequencePList.rend();
    bool noSequence=false;
    if (aSequencePListRI==aSequencePListRend)
      // we don't have any sequence left, meaning there is either no sequence at all 
      // or we are at the end where there is a bit left without a sequence
      noSequence=true;
    bool done=false; // are we all done?
    while (!done) { 
      // deal with the bit outside the current Sequence
      while (noSequence ||
	     (*aSequencePListRI)->myLastElement_p!=(*aBasicBlockElementListRI)) { 
	// let the stuff outside the sequence insert itself 
	dynamic_cast<BasicBlockElementAlg&>((**aBasicBlockElementListRI).getBasicBlockElementAlgBase()).insertYourself(getContaining());
	++aBasicBlockElementListRI;
	if (aBasicBlockElementListRI==getContaining().getBasicBlockElementList().rend()) { 
	  done=true;
	  break;
	}
      } // end while
	// now we are done or (*aSequencePListRI)->myLastElement_p==(*aBasicBlockElementListRI))
      while (!done  // this implies noSequence=false
	     && 
	     (*aSequencePListRI)->myFirstElement_p!=(*aBasicBlockElementListRI)) { 
	// this should bring us to the begin of the sequence
	++aBasicBlockElementListRI;
	if (aBasicBlockElementListRI==getContaining().getBasicBlockElementList().rend()) { 
	  done=true;
	  break;
	}
      }
      if (!done) { 
	const xaifBoosterDerivativePropagator::DerivativePropagator& aDerivativePropagator((*aSequencePListRI)->myDerivativePropagator);
	for(xaifBoosterDerivativePropagator::DerivativePropagator::EntryPList::const_reverse_iterator entryPListI=
	      aDerivativePropagator.getEntryPList().rbegin();
	    entryPListI!= aDerivativePropagator.getEntryPList().rend();
	    ++entryPListI) {
	  reinterpretDerivativePropagatorEntry(**entryPListI);
	} // end for DerivativePropagatorEntry list
	++aSequencePListRI;
	if (aSequencePListRI==aSequencePListRend)
	  // there is no sequence left
	  noSequence=true;
      } // end of if (!done)
    } // end while (!done)
  } 

  void BasicBlockAlg::addZeroDeriv(Variable& theTarget,
				   const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theSubroutineCall(addInlinableSubroutineCall("ZeroDeriv",
														 aReversalType));
    theSubroutineCall.setId("inline_zeroderiv");
    theTarget.copyMyselfInto(theSubroutineCall.addConcreteArgument(1).getArgument().getVariable());
  } 

  void BasicBlockAlg::addUnitFactor(Variable& theSource,
				    Variable& theTarget,
				    const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    /** 
     * The inlinable call incDeriv(y,x) is supposed to do x.d+=y.d
     * HOWEVER, this should not happen when y and x are the same
     * so we make an educated guess based on the alias information
     * and add a checker if we can't decide. 
     */
    if (ConceptuallyStaticInstances::instance()->getCallGraph().getAliasMap().mayAlias(theTarget.getAliasMapKey(),
										       theSource.getAliasMapKey())) { 
      xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theSetDerivCall(addInlinableSubroutineCall("CondIncZeroDeriv",
														 aReversalType));
      theSetDerivCall.setId("inline_CondIncZeroDeriv");
      theTarget.copyMyselfInto(theSetDerivCall.addConcreteArgument(1).getArgument().getVariable());
      theSource.copyMyselfInto(theSetDerivCall.addConcreteArgument(2).getArgument().getVariable());
    }
    else { 
      xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theSetDerivCall(addInlinableSubroutineCall("IncDeriv",
														 aReversalType));
      theSetDerivCall.setId("inline_IncDeriv");
      theTarget.copyMyselfInto(theSetDerivCall.addConcreteArgument(1).getArgument().getVariable());
      theSource.copyMyselfInto(theSetDerivCall.addConcreteArgument(2).getArgument().getVariable());
      xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theZeroDerivCall(addInlinableSubroutineCall("ZeroDeriv",
														  aReversalType));
      theZeroDerivCall.setId("inline_zeroderiv");
      theTarget.copyMyselfInto(theZeroDerivCall.addConcreteArgument(1).getArgument().getVariable());
    }
  } 

  const Variable& BasicBlockAlg::addFactorPop(const Symbol& aTemporarySymbol,
					      const Scope& theScopeOfTheTemporarySymbol,
					      const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theZeroDerivCall(addInlinableSubroutineCall("Pop",
														aReversalType));
    theZeroDerivCall.setId("inline_pop");
    Variable& theInlineVariable(theZeroDerivCall.addConcreteArgument(1).getArgument().getVariable());
    // give it a name etc.
    // create a new symbol and add a new VariableSymbolReference in the Variable
    VariableSymbolReference* theNewVariableSymbolReference_p=
      new VariableSymbolReference(aTemporarySymbol,
				  getContaining().getScope());
    theNewVariableSymbolReference_p->setId("1");
    theNewVariableSymbolReference_p->setAnnotation("xaifBoosterBasicBlockPreaccumulationTapeAdjoint::BasicBlockAlg::addFactorPop");
    // pass it on to the variable and relinquish ownership
    theInlineVariable.supplyAndAddVertexInstance(*theNewVariableSymbolReference_p);
    theInlineVariable.getAliasMapKey().setTemporary();
    theInlineVariable.getDuUdMapKey().setTemporary();
    return theInlineVariable;
  } 

  void BasicBlockAlg::addSaxpy(const Variable& theSource,
			       const Variable& theTarget,
			       const Variable& theFactor,
			       const xaifBoosterDerivativePropagator::DerivativePropagatorEntry& aDerivativePropagatorEntry,
			       const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theSaxpyCall(addInlinableSubroutineCall("Saxpy",
													    aReversalType));
    theSaxpyCall.setId("inline_saxpy");
    theFactor.copyMyselfInto(theSaxpyCall.addConcreteArgument(1).getArgument().getVariable());
    theTarget.copyMyselfInto(theSaxpyCall.addConcreteArgument(2).getArgument().getVariable());
    theSource.copyMyselfInto(theSaxpyCall.addConcreteArgument(3).getArgument().getVariable());
    if (!aDerivativePropagatorEntry.isIncremental()) { 
      xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theZeroDerivCall(addInlinableSubroutineCall("ZeroDeriv",
														  aReversalType));
      theZeroDerivCall.setId("inline_zeroderiv");
      theTarget.copyMyselfInto(theZeroDerivCall.addConcreteArgument(1).getArgument().getVariable());
    }
  }

  void BasicBlockAlg::addSaxpy_constantFactor(const Variable& theSource,
					      const Variable& theTarget,
					      const Constant& theConstantFactor,
					      const xaifBoosterDerivativePropagator::DerivativePropagatorEntry& aDerivativePropagatorEntry,
					      const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theSaxpyCall (addInlinableSubroutineCall("Saxpy",
													     aReversalType));
    theSaxpyCall.setId("inline_saxpy");

    ConcreteArgument& theNewArgument (theSaxpyCall.addConcreteArgument(1));
    theNewArgument.makeConstant(theConstantFactor.getType());
    theNewArgument.getConstant().setFromString(theConstantFactor.toString());

    theTarget.copyMyselfInto(theSaxpyCall.addConcreteArgument(2).getArgument().getVariable());
    theSource.copyMyselfInto(theSaxpyCall.addConcreteArgument(3).getArgument().getVariable());
    if (!aDerivativePropagatorEntry.isIncremental()) { 
      xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& theZeroDerivCall(addInlinableSubroutineCall("ZeroDeriv",
														  aReversalType));
      theZeroDerivCall.setId("inline_zeroderiv");
      theTarget.copyMyselfInto(theZeroDerivCall.addConcreteArgument(1).getArgument().getVariable());
    }
  } // end BasicBlockAlg::addSaxpy_constantFactor()

  void BasicBlockAlg::reinterpretDerivativePropagatorEntry(const xaifBoosterDerivativePropagator::DerivativePropagatorEntry& aDerivativePropagatorEntry) { 
    // retrieve stored index values if needed starting with the target: 
    // take care of target address if needed: 
    const Variable& theOriginalTarget(aDerivativePropagatorEntry.getTarget());
    // this is either the original target
    // or it is the replaced target in case of ArrayAccesses
    Variable theActualTargetAnonymous,theActualTargetExplicit;
    reinterpretArrayAccess(theOriginalTarget,theActualTargetAnonymous,ForLoopReversalType::ANONYMOUS); 
    reinterpretArrayAccess(theOriginalTarget,theActualTargetExplicit,ForLoopReversalType::EXPLICIT); 
    xaifBoosterDerivativePropagator::DerivativePropagatorEntry::FactorList aFactorList;
    aDerivativePropagatorEntry.getFactors(aFactorList);
    for (xaifBoosterDerivativePropagator::DerivativePropagatorEntry::FactorList::reverse_iterator aFactorListI=aFactorList.rbegin();
	 aFactorListI!=aFactorList.rend();
	 ++aFactorListI) { 
      if ((*aFactorListI).getKind()==xaifBoosterDerivativePropagator::DerivativePropagatorEntry::Factor::ZERO_FACTOR) { 
	addZeroDeriv(theActualTargetAnonymous,ForLoopReversalType::ANONYMOUS);
	addZeroDeriv(theActualTargetExplicit,ForLoopReversalType::EXPLICIT);
	break; 
      } // ZERO_FACTOR
      else { 
	const Variable& theOriginalSource((*aFactorListI).getSource());
	// this is either the original source
	// or it is the replaced target in case of ArrayAccesses
	Variable theActualSourceAnonymous,theActualSourceExplicit;
	reinterpretArrayAccess(theOriginalSource,theActualSourceAnonymous,ForLoopReversalType::ANONYMOUS); 
	reinterpretArrayAccess(theOriginalSource,theActualSourceExplicit,ForLoopReversalType::EXPLICIT); 
	// deal with the other cases: 
	switch((*aFactorListI).getKind()) { 
	case xaifBoosterDerivativePropagator::DerivativePropagatorEntry::Factor::UNIT_FACTOR: // the SetDeriv in tlm
	  { 
	    addUnitFactor(theActualSourceAnonymous,theActualTargetAnonymous,ForLoopReversalType::ANONYMOUS);
	    addUnitFactor(theActualSourceExplicit,theActualTargetExplicit,ForLoopReversalType::EXPLICIT);
	    break; 
	  } 
	case xaifBoosterDerivativePropagator::DerivativePropagatorEntry::Factor::CONSTANT_FACTOR: {
	  /**
	   * The inlinable call saxpy(a,x,y) is supposed to do y.d=y.d+a*x.d which in reverse means
	   * we have to switch arguments properly
	   */
	  // do the anonymous version
	  addSaxpy_constantFactor(theActualSourceAnonymous,
				  theActualTargetAnonymous,
				  (*aFactorListI).getConstant(),
				  aDerivativePropagatorEntry,
				  ForLoopReversalType::ANONYMOUS);
	  // do the explicit version
	  addSaxpy_constantFactor(theActualSourceExplicit,
				  theActualTargetExplicit,
				  (*aFactorListI).getConstant(),
				  aDerivativePropagatorEntry,
				  ForLoopReversalType::EXPLICIT);
	  break;
	} // end case CONSTANT_FACTOR
	case xaifBoosterDerivativePropagator::DerivativePropagatorEntry::Factor::VARIABLE_FACTOR: // both represent some saxpy or sax 
	  { 
	    // the inlinable call needs a temporary which contains the factor
	    const Symbol& aTemporarySymbol(getContaining().getScope().
					   getSymbolTable().
					   addUniqueAuxSymbol(SymbolKind::VARIABLE,
							      SymbolType::REAL_STYPE,
							      SymbolShape::SCALAR,
							      false));
	    // do the anonymous version
	    const Variable& theFactorVariable (addFactorPop(aTemporarySymbol,
							    getContaining().getScope(),			    
							    ForLoopReversalType::ANONYMOUS));
	    // do the explicit version - we can of course use the same symbol so 
	    // theFactorVariable_p is identical
	    addFactorPop(aTemporarySymbol,
			 getContaining().getScope(),			    
			 ForLoopReversalType::EXPLICIT);
	    /** 
	     * The inlinable call saxpy(a,x,y) is supposed to do y.d=y.d+a*x.d which in reverse means
	     * we have to switch arguments properly
	     */
	    // do the anonymous version
	    addSaxpy(theActualSourceAnonymous,
		     theActualTargetAnonymous,
		     theFactorVariable,
		     aDerivativePropagatorEntry,
		     ForLoopReversalType::ANONYMOUS);
	    // do the explicit version
	    addSaxpy(theActualSourceExplicit,
		     theActualTargetExplicit,
		     theFactorVariable,
		     aDerivativePropagatorEntry,
		     ForLoopReversalType::EXPLICIT);
	    break; 
	  } 
	default: 
	  THROW_LOGICEXCEPTION_MACRO("BasicBlockAlg::reinterpretDerivativePropagatorEntry: cannot handle factor kind " 
				     << (*aFactorListI).getKind()); 
	  break; 
	} // end switch 
      } // end else (other versions)
    } // end for FactorList
  }

  const PlainBasicBlock::BasicBlockElementList& 
  BasicBlockAlg::getBasicBlockElementList(const ForLoopReversalType::ForLoopReversalType_E& aReversalType) const { 
    switch(aReversalType) { 
    case ForLoopReversalType::ANONYMOUS : 
      return myBasicBlockElementListAnonymousReversal;
      break;
    case ForLoopReversalType::EXPLICIT : 
      return myBasicBlockElementListExplicitReversal;
      break;
    default: 
      THROW_LOGICEXCEPTION_MACRO("BasicBlockAlg::getBasicBlockElementList: unknown reversal type "
				 << ForLoopReversalType::toString(aReversalType).c_str());
      break;
    }      
    // to make the compiler happy:
    return myBasicBlockElementListAnonymousReversal;
  }  

  PlainBasicBlock::BasicBlockElementList& 
  BasicBlockAlg::getBasicBlockElementList(const ForLoopReversalType::ForLoopReversalType_E& aReversalType) { 
    switch(aReversalType) { 
    case ForLoopReversalType::ANONYMOUS : 
      return myBasicBlockElementListAnonymousReversal;
      break;
    case ForLoopReversalType::EXPLICIT : 
      return myBasicBlockElementListExplicitReversal;
      break;
    default: 
      THROW_LOGICEXCEPTION_MACRO("BasicBlockAlg::getBasicBlockElementList: unknown reversal type "
				 << ForLoopReversalType::toString(aReversalType).c_str());
      break;
    }      
    // to make the compiler happy:
    return myBasicBlockElementListAnonymousReversal;
  }  

} // end namespace xaifBoosterBasicBlockPreaccumulationTapeAdjoint

